from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem


# Standard library imports
import numpy as np
from skimage.measure import marching_cubes
import scipy as sp

import plotly.graph_objects as go


def plot_surface_plotly(
        self: 'BaseSystem',
        field: np.ndarray,
        value: float,
        alpha: float = 0.5,
        color: str = 'b',
        plt_colormap_object: Any = None,
        hovertemplate: str = None,
        customdata: np.ndarray = None,
        **kwargs: Any
        ) -> go.Mesh3d:
    """Plot the surface of the given field.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    field : ndarray
        3D array containing the field values
    value : float
        Isosurface value
    alpha : float, optional
        Transparency value. Defaults to 0.5.
    color : str, optional
        Surface color. Defaults to 'b'.
    hovertemplate : str
        Template for the hover text.
    kwargs : Any
        Plotting keyword arguments.

    Returns
    -------
    go.Mesh3D
        The plotly mesh object containing the surface plot.
    """

    customdata = None

    verts, faces, _, _ = marching_cubes(field, value)
    
    # Calculate the centroids of each triangle
    centroids = np.mean(verts[faces], axis=1)

    # Color according to the angle
    if np.iscomplexobj(color):
        x, y, z = np.mgrid[0:field.shape[0], 0:field.shape[1], 0:field.shape[2]]
        points = np.c_[x.ravel(), y.ravel(), z.ravel()]

        reals = np.real(color)
        imags = np.imag(color)

        interpolation_method = kwargs.get('interpolation_method', 'linear')
        # print("Interpolating points with method ", interpolation_method, ".")
        # print("If this process is slow, consider passing 'interpolation_method='nearest' with the plot_complex_field function.")
        # print("That will speed up the process, but the plot may look less smooth.")
        reals_faces = sp.interpolate.griddata(points, reals.ravel(), centroids, method='nearest')
        imags_faces = sp.interpolate.griddata(points, imags.ravel(), centroids, method='nearest')
        # print("Interpolation done.")

        theta_faces = np.arctan2(imags_faces, reals_faces)

        # Normalize theta values for color mapping
        theta_faces_normalized = (theta_faces + np.pi) / (2*np.pi)

        # Map normalized theta values to colors
        colors = plt_colormap_object(theta_faces_normalized)

        

        theta_faces = np.arctan2(imags_faces, reals_faces)

        hovertemplate=kwargs['xlabel']+': %{x:.2f}<br>'+\
                        kwargs['ylabel']+': %{y:.2f}<br>'+\
                        kwargs['zlabel']+': %{z:.2f}<br>'+\
                        'amplitude: '+ f'{value:.2e}<br>'+\
                        'phase: %{customdata:.2f} π'

        # hovertemplate = kwargs['xlabel']+': %{x:.2f}<br>'+\
        #                 kwargs['ylabel']+': %{y:.2f}<br>'+\
        #                 kwargs['zlabel']+': %{z:.2f}<br>'+\
        #                 'field: '+ f'{value:.2e}'

        customdata = theta_faces/np.pi
        
        # Convert colors to 'rgba()' format required by Plotly
        colors = ['rgba({},{},{},{})'.format(*c) for c in (255*colors).astype(int)]
        color = None

    elif isinstance(alpha,np.ndarray):

        alpha_max = np.max(alpha)
        alpha_min = np.min(alpha)

        max_opacity = 0.9

        if (alpha_max - alpha_min)<1e-6:
            alpha = max_opacity*np.ones_like(alpha)
        else:
            alpha = max_opacity*(alpha - alpha_min) / (alpha_max - alpha_min)

        x, y, z = np.mgrid[0:field.shape[0], 0:field.shape[1], 0:field.shape[2]]
        points = np.c_[x.ravel(), y.ravel(), z.ravel()]

        interpolation_method = kwargs.get('interpolation_method', 'linear')
        valid_centroids = ~np.isnan(centroids).any(axis=1)
        rho_normalized_faces = np.full(centroids.shape[0], np.nan)
        rho_normalized_faces[valid_centroids] = sp.interpolate.griddata(points, alpha.ravel(), centroids[valid_centroids], method='nearest')

        colors =   ['rgba({},{},{},{})'.format(*(np.array([
                    (255*color[0]).astype(int),
                    (255*color[1]).astype(int),
                    (255*color[2]).astype(int),
                    0 if np.isnan(rho_normalized_faces[i]) else rho_normalized_faces[i]
                    ]))) for i in range(len(faces))]
        
        hovertemplate=kwargs['xlabel']+': %{x:.2f}<br>'+\
                        kwargs['ylabel']+': %{y:.2f}<br>'+\
                        kwargs['zlabel']+': %{z:.2f}<br>'+\
                        'amplitude: %{customdata:.2e}<br>' + \
                        'phase: '+ f'{value/np.pi:.2f} π '
        
        customdata = alpha_min + rho_normalized_faces*(alpha_max-alpha_min)

        color = 'rgba({},{},{},{})'.format(*(255*np.array(color)).astype(int)) 
        # color = None
        alpha = None

    else:

        colors = None
        color = 'rgba({},{},{},{})'.format(*(255*np.array(color)).astype(int)) 


    # Extract coordinates
    x = kwargs.get('x', self.x/self.a0).flatten()
    dx = x[1] - x[0]
    xmin = x[0]
    xmax = x[-1]+dx

    y = kwargs.get('y', self.y/self.a0).flatten()
    dy = y[1] - y[0]
    ymin = y[0]
    ymax = y[-1]+dy

    z = kwargs.get('z', self.z/self.a0).flatten()
    dz = z[1] - z[0]
    zmin = z[0]
    zmax = z[-1]+dz

    # Create the mesh object
    x_new = (verts[:, 0] * dx + xmin) 
    y_new = (verts[:, 1] * dy + ymin) 
    z_new = (verts[:, 2] * dz + zmin) 


    if hovertemplate is None:
        hovertemplate = kwargs['xlabel']+': %{x:.2f}<br>'+\
                        kwargs['ylabel']+': %{y:.2f}<br>'+\
                        kwargs['zlabel']+': %{z:.2f}<br>'+\
                        'field: '+ f'{value:.2e}'


    mesh = go.Mesh3d(
                x=x_new, 
                y=y_new, 
                z=z_new,
                i=faces[:, 0], 
                j=faces[:, 1], 
                k=faces[:, 2],
                facecolor=colors,  # Set color for each face
                color=color,
                showscale=False,
                opacity=alpha,
                hovertemplate=hovertemplate,
                customdata=customdata,
                name='',
                scene=kwargs['ax']['sceneN'],
                hoverlabel=dict(
                    bgcolor=color
                )
            )

    plot_shadows = kwargs.get('plot_shadows', True)
    if plot_shadows:
        shadow_meshes = []

        # Shadow on xy-plane:
        mesh_shadow_xy = go.Mesh3d(
            x=x_new, 
            y=y_new, 
            z=0*z_new + zmin,
            i=faces[:, 0], 
            j=faces[:, 1], 
            k=faces[:, 2],
            color='black',
            showscale=False,
            opacity=0.2,
            name='',
            showlegend=False,
            scene=kwargs['ax']['sceneN']
            )

        # Shadow on xz-plane:
        mesh_shadow_xz = go.Mesh3d(
            x=x_new, 
            y=0*y_new + ymin, 
            z=z_new,
            i=faces[:, 0], 
            j=faces[:, 1], 
            k=faces[:, 2],
            color='black',
            showscale=False,
            opacity=0.2,
            name='',
            showlegend=False,
            scene=kwargs['ax']['sceneN']
            )

        # Shadow on yz-plane:
        mesh_shadow_yz = go.Mesh3d(
            x=0*x_new + xmin, 
            y=y_new, 
            z=z_new,
            i=faces[:, 0], 
            j=faces[:, 1], 
            k=faces[:, 2],
            color='black',
            showscale=False,
            opacity=0.2,
            name='',
            showlegend=False,
            scene=kwargs['ax']['sceneN']
            )

        shadow_meshes.append(mesh_shadow_xy)
        shadow_meshes.append(mesh_shadow_xz)
        shadow_meshes.append(mesh_shadow_yz)
    else:
        shadow_meshes = None

    return mesh, shadow_meshes


from comfit.phase_field_crystal.phase_field_crystal import PhaseFieldCrystal

import numpy as np
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint

from comfit.tool.tool_print_in_color import tool_print_in_color

class PhaseFieldCrystal3DSimpleCubic(PhaseFieldCrystal):
    def __init__(self, nx, ny, nz, **kwargs):
        """
        Initializes a phase field crystal system in 3D with a simple cubic crystal structure.

        Parameters
        ----------
        nx : int
            The number of unit cells in the x direction.
        ny : int
            The number of unit cells in the y direction.
        nz : int
            The number of unit cells in the z direction.
        kwargs : dict
            Additional arguments to set as attributes.

        Returns
        -------
        None
        """
        # Type of the system
        self.type = 'PhaseFieldCrystal3DSimpleCubic'
        self.dim = 3
        self.nx = nx
        self.ny = ny
        self.nz = nz

        # Default simulation parameters
        self.micro_resolution = [5, 5, 5]
        self.psi0 = -0.325
        self.r = -0.3
        self.t = 0
        self.v = 1
        self.dt = 0.1
        # TODO: Some of this code seems to be reprinted. Maybe it should be moved to the BaseSystem class? (Vidar 18.12.23)

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.xRes = nx * self.micro_resolution[0]
        self.yRes = ny * self.micro_resolution[1]
        self.zRes = nz * self.micro_resolution[2]

        a0 = 2 * np.pi
        self.a = a0 * np.array([[1, 0, 0],
                                [0, 1, 0],
                                [0, 0, 1]])

        self.q = np.array([[1, 0, 0],
                           [0, 1, 0],
                           [0, 0, 1],
                           [0, 1, 1],
                           [1, 0, 1],
                           [1, 1, 0],
                           [0, -1, 1],
                           [-1, 0, 1],
                           [-1, 1, 0],
                           [-1, 1, 1],
                           [1, -1, 1],
                           [1, 1, -1],
                           [1, 1, 1]], dtype=float)

        # Set the number of reciprocal modes
        self.number_of_reciprocal_lattice_modes = 13
        self.number_of_primary_reciprocal_lattice_modes = 3

        # Set the grid
        self.dx = a0 / self.micro_resolution[0]
        self.dy = a0 / self.micro_resolution[1]
        self.dz = a0 / self.micro_resolution[2]

        self.type_of_evolution = kwargs.get('type_of_evolution', 'conserved')
        self.A_proto, self.B_proto, self.C_proto = self.calc_proto_amplitudes_conserved()

        # Set the elastic constants
        self.el_lambda_proto = 16 * self.B_proto ** 2 + 128 * self.C_proto ** 2
        self.el_mu_proto = 16 * self.B_proto ** 2 + 128 * self.C_proto ** 2
        self.el_gamma_proto = 32*self.A_proto**2 - 16*self.B_proto**2 - 256*self.C_proto**2

        bool_is_for_properties_calculation = kwargs.get('for_properties_calculation', False)

        if not bool_is_for_properties_calculation:
            tool_print_in_color('Initiating a 3D simple cubic PFC model.', 'green')
            kwargs.pop('for_properties_calculation', None)
            pfc = PhaseFieldCrystal3DSimpleCubic(1,1,1,for_properties_calculation=True, type_of_evolution=self.type_of_evolution, **kwargs)
            final_strain, self.psi0, self.A, self.B, self.C, self.el_lambda, self.el_mu, self.el_gamma = pfc.calc_strained_amplitudes()     
        else:
            self.A = self.A_proto
            self.B = self.B_proto
            self.C = self.C_proto
            self.el_lambda = self.el_lambda_proto
            self.el_mu = self.el_mu_proto
            self.el_gamma = self.el_gamma_proto


        self.eta0 = np.array([self.A, self.A, self.A,
                     self.B, self.B, self.B, self.B, self.B, self.B,
                     self.C, self.C, self.C, self.C])

        # Initialize the BaseSystem
        kwargs['xRes'] = self.xRes
        kwargs['yRes'] = self.yRes
        kwargs['zRes'] = self.zRes
        kwargs['dx'] = self.dx
        kwargs['dy'] = self.dy
        kwargs['dz'] = self.dz
        kwargs['dt'] = self.dt
        super().__init__(self.dim,  **kwargs)

        # Set the a0
        self.a0 = a0
        self.bool_has_defined_length_scale = True

        if not bool_is_for_properties_calculation:
            self.conf_apply_distortion([[final_strain,0,0],[0,final_strain,0],[0,0,final_strain]], update_q_and_a_vectors=True)
            self.a0 = self.a0 * (1+final_strain)

    def calc_free_energy_from_proto_amplitudes(self, psi0, A, B, C):
        """
        Calculates the free energy of the system from the proto-amplitudes.

        Parameters
        ----------
        psi0 : float
            The value of psi0.
        A : float
            The value of A.
        B : float
            The value of B.
        C : float
            The value of C.

        Returns
        -------
        float
            The free energy of the system.
        """
        r = self.r
        t = self.t
        v = self.v
        return 2*np.pi**3/3*(48*C**2*r + 270*A**4*v + 1620*B**4*v + 576*A**3*C*v + 648*C**4*v + 96*C**2*t*psi0 + 6*(36 + r + 24*C**2*v)*psi0**2 + 4*t*psi0**3 + 3*v*psi0**4 + 192*B**3*(t + 3*v*psi0) + 576*A*B*C*(t + 3*v*(3*B + psi0)) + 36*A**2*( r + 96*B**2*v + 36*C**2*v + 2*t*psi0 + 3*v*psi0**2 + 8*B*( t + 3*v*psi0)) + 72*B**2*(r + 54*C**2*v + psi0*(2*t + 3*v*psi0)) )

    def calc_proto_amplitudes_conserved(self):
        """
        Calculates the proto-amplitudes for the system.

        Parameters
        ----------
        None

        Returns
        -------
        tuple
            The proto-amplitudes for the system.
        """
        psi0 = self.psi0
        r = self.r
        t = self.t
        v = self.v

        A = 0
        B = 0
        C = 0
        free_energy = self.calc_free_energy_from_proto_amplitudes(psi0, A, B, C)

        for A0 in np.linspace(0.01, 1, 10):
            for B0 in [A0/2]:
                for C0 in [A0/4]:
                    [A_tmp, B_tmp, C_tmp] = fsolve(self.calc_proto_amplitude_equations_conserved, [A0, B0, C0])
                    free_energy_tmp = self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp, B_tmp, C_tmp)

                    if free_energy_tmp <= free_energy:
                        A = A_tmp
                        B = B_tmp
                        C = C_tmp
                        free_energy = free_energy_tmp

        return A, B, C

    def calc_proto_amplitude_equations_conserved(self, vars):
        """
        Calculates the equations for the proto-amplitudes for the system in the case of conserved dynamics.

        Parameters
        ----------
        vars : list
            The proto-amplitudes for the system.

        Returns
        -------
        list
            The equations for the proto-amplitudes for the system in the case of conserved dynamics.
        """
        psi0 = self.psi0
        r = self.r
        t = self.t
        v = self.v

        A, B, C = vars

        eq1 = 15*A**3*v + 24*A**2*C*v + 8*B*C*(t + 3*v*(3*B + psi0)) + A*( r + 96*B**2*v + 36*C**2*v + 2*t*psi0 + 3*v*psi0**2 + 8*B*(t + 3*v*psi0))
        eq2 = 45*B**3*v + 4*B**2*(t + 3*v*psi0) + 2*A*(A + 2*C)*(t + 3*v*psi0) + B*( r + 48*A**2*v + 72*A*C*v + 54*C**2*v + 2*t*psi0 + 3*v*psi0**2)
        eq3 = 27*C**3*v + C*( r + 27*A**2*v + 81*B**2*v + 2*t*psi0 + 3*v*psi0**2) + 6*A*(A**2*v + 9*B**2*v + B*(t + 3*v*psi0))

        return [eq1, eq2, eq3]

    def calc_L_f(self):
        """
        Calculates the L operator in Fourier space.

        Parameters
        ----------
        None

        Returns
        -------
        numpy.ndarray
            The L operator in Fourier space.
        """
        k2 = self.calc_k2()
        return (1 - k2)*(2 - k2)*(3 - k2)

    def calc_L_sum_f(self):
        """
        Calculates the sum of the L operators in Fourier space. Needed for stress calculation functions.

        Parameters
        ----------
        None

        Returns
        -------
        numpy.ndarray
            The L operator in Fourier space.
        """
        k2 = self.calc_k2()
        return 11 - 12*k2 + 3*k2**2


from typing import TYPE_CHECKING, Any, Dict, Tuple

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
import numpy as np
import scipy as sp

# Third-party library imports
import plotly.graph_objects as go
import matplotlib.colors as mcolors
import matplotlib.cm as cm
from skimage.measure import marching_cubes

from .plot_surface_plotly import plot_surface_plotly

# Local application imports
from comfit.tool import (
    tool_complete_field,
    tool_colormap, 
    tool_set_plot_axis_properties_plotly,
    tool_plotly_colorbar,
    tool_plotly_define_2D_plot_ax,
    tool_plotly_define_3D_plot_ax
)


def plot_complex_field_plotly(
        self: 'BaseSystem',
        complex_field: np.ndarray,
        **kwargs: Any
        ) -> Tuple[go.Figure, Dict]:
    """Plot a complex field using Plotly.

    Creates a visualization of a complex field using various plotting methods
    depending on the dimensionality of the field.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    complex_field : np.ndarray
        The complex field to plot
    kwargs : Any
        Additional keyword arguments to customize the plot

    Returns
    -------
    go.Figure
        The Plotly figure containing the plot
    """

    kwargs['colormap'] = kwargs.get('colormap', 'angle') # Override the default colormap with 'angle'
    complex_field, fig, ax, kwargs = self.plot_prepare(complex_field, field_type = 'complex', **kwargs)

    # Kewyord arguments
    kwargs['colorbar'] = kwargs.get('colorbar', True)

    # Extend the field if not a complete array is given
    complex_field = tool_complete_field(self, complex_field)

    # Calculate the magnitude and phase of the complex field
    rho = np.abs(complex_field)
    theta = np.angle(complex_field)

    plt_colormap_object = tool_colormap(kwargs['colormap'], plot_lib='matplotlib')

    ###############################################################
    ###################### DIMENSION: 1 ###########################
    ###############################################################

    if self.dim == 1:

        ax = tool_plotly_define_2D_plot_ax(fig, ax)

        vlim = kwargs.get('vlim', None)
        if vlim is not None:
            kwargs['ylim'] = vlim

        # Color in the graph based on the argument of the complex field
        blend_factor=0.3 # The degree to which the color is blended with white
        
        color = (1-blend_factor)*np.array(plt_colormap_object((theta[0] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
        color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')

        # Extract coordinates
        x = kwargs.get('x', self.x/self.a0).flatten()
        dx = x[1] - x[0]
        xmin = x[0]
        xmax = x[-1]+dx

        fig.add_trace(go.Scatter(x=[xmin,(xmin+dx/2)], 
                        y=[rho[0],(rho[0]+rho[1])/2],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str,
                        xaxis=ax['xN'],
                        yaxis=ax['yN']))

        for i in range(1,self.xRes-1):
            color = (1-blend_factor)*np.array(plt_colormap_object((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
            color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
            fig.add_trace(go.Scatter(x=[(x[i]-dx/2),x[i]], 
                        y=[(rho[i]+rho[i-1])/2,rho[i]],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str, 
                        xaxis=ax['xN'],
                        yaxis=ax['yN']))

            color = (1-blend_factor)*np.array(plt_colormap_object((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
            color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
            fig.add_trace(go.Scatter(x=[x[i],(x[i]+dx/2)], 
                        y=[rho[i],(rho[i]+rho[i+1])/2],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str,
                        xaxis=ax['xN'],
                        yaxis=ax['yN']))

        color = (1-blend_factor)*np.array(plt_colormap_object((theta[-1] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
        color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
        fig.add_trace(go.Scatter(x=[(xmax-1.5*dx),(xmax-dx)], 
                        y=[(rho[-1]+rho[-2])/2,rho[-1]],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str,
                        xaxis=ax['xN'],
                        yaxis=ax['yN']))


        fig.add_trace(go.Scatter(
            x=x,
            y=rho,
            mode='lines',
            showlegend=False,
            customdata=np.stack((theta/np.pi, rho), axis=-1),
            hovertemplate= kwargs['xlabel']+': %{x:.2f} <br>'+\
                                'amplitude: %{customdata[1]:.2e}<br>'+\
                                'phase: %{customdata[0]:.2f} π',
            name='',
            line=dict(color='black'),
            xaxis=ax['xN'],
            yaxis=ax['yN']
        ))

        
    ###############################################################
    ###################### DIMENSION: 2 ###########################
    ###############################################################

    elif self.dim == 2:
        
        # Keyword arguments particular to the 2D case
        plot_method = kwargs.get('plot_method', 'phase_angle')

        # Extract coordinates
        x = kwargs.get('x', self.x/self.a0).flatten()
        dx = x[1] - x[0]
        xmin = x[0]
        xmax = x[-1]+dx
        
        y = kwargs.get('y', self.y/self.a0).flatten()
        dy = y[1] - y[0]
        ymin = y[0]
        ymax = y[-1]+dy

        rho_normalized = rho / np.max(rho)

        if plot_method == '3Dsurface':

            ax = tool_plotly_define_3D_plot_ax(fig, ax)

            # Plot field as 3D surface where height is amplitude and color is phase
            X, Y = np.meshgrid(x, y, indexing='ij')

            surface = go.Surface(
                x=X,
                y=Y,
                z=rho,
                surfacecolor=theta,
                colorscale=kwargs['colormap_object'],
                customdata=np.stack((theta/np.pi, rho), axis=-1),
                hovertemplate=(f"{kwargs['xlabel']}: %{{x:.2f}}<br>" +
                              f"{kwargs['ylabel']}: %{{y:.2f}}<br>" +
                              "amplitude: %{customdata[1]:.2e}<br>" +
                              "phase: %{customdata[0]:.2f} π"),
                showscale=False
            )

            fig.add_trace(surface)


        elif plot_method == 'phase_angle':
            
            ax = tool_plotly_define_2D_plot_ax(fig, ax)
            
            norm = mcolors.Normalize(vmin=-np.pi, vmax=np.pi)

            # Create an RGBA image array
            rgba_image = cm.ScalarMappable(norm=norm, cmap=plt_colormap_object).to_rgba(theta)

            # Apply the spatial opacity
            for i in range(rgba_image.shape[0]):
                for j in range(rgba_image.shape[1]):
                    for k in range(3):
                        rgba_image[i, j, k] = 1-(1-rgba_image[i, j, k])*rho_normalized[i, j]  # Set alpha channel

            rgba_image = np.transpose(rgba_image, (1, 0, 2))

            # Convert RGBA image to a format Plotly can understand
            image_data = (rgba_image * 255).astype(np.uint8)

            trace = go.Image(z=image_data, 
                            dx=dx, 
                            dy=dy, 
                            x0=xmin, 
                            y0=ymin,
                            hovertemplate=kwargs['xlabel']+': %{x:.2f}<br>'+\
                                            kwargs['ylabel']+': %{y:.2f}<br>'+\
                                            'amplitude: %{customdata[1]:.2e}<br>'+\
                                            'phase: %{customdata[0]:.2f} π',
                            customdata=np.stack((np.transpose(theta/np.pi), np.transpose(rho)), axis=-1),
                            name='',
                            xaxis=ax['xN'],
                            yaxis=ax['yN']
                            )

            fig.add_trace(trace)

            # Add grid lines manually since the default plotly grid lines are
            # not available for go.Image traces
            if kwargs['grid']:
                N = 11
                # Draw N equally spaced horizontal and vertical lines
                x_positions = np.linspace(x[0], x[-1], N)
                y_positions = np.linspace(y[0], y[-1], N)

                # Add horizontal grid lines
                for y_pos in y_positions:
                    fig.add_trace(go.Scatter(
                        x=[xmin, xmax],
                        y=[y_pos, y_pos],
                        mode='lines',
                        line=dict(color='grey', width=0.5),
                        showlegend=False,
                        hoverinfo='skip',
                        xaxis=ax['xN'],
                        yaxis=ax['yN']
                    ))

                # Add vertical grid lines
                for x_pos in x_positions:
                    fig.add_trace(go.Scatter(
                        x=[x_pos, x_pos],
                        y=[ymin, ymax],
                        mode='lines',
                        line=dict(color='grey', width=0.5),
                        showlegend=False,
                        hoverinfo='skip',
                        xaxis=ax['xN'],
                        yaxis=ax['yN']
                    ))


    ###############################################################
    ###################### DIMENSION: 3 ###########################
    ###############################################################
    elif self.dim == 3:

        ax = tool_plotly_define_3D_plot_ax(fig, ax)
        kwargs['ax'] = ax 

        # Keyword arguments particular to the 3D case
        plot_method = kwargs.get('plot_method', 'phase_blob')
        
        rho_normalized = rho / np.max(rho)

        if plot_method == 'phase_angle':

            # Extract coordinates
            x = kwargs.get('x', self.x/self.a0).flatten()
            dx = x[1] - x[0]
            xmin = x[0]
            xmax = x[-1]+dx

            y = kwargs.get('y', self.y/self.a0).flatten()
            dy = y[1] - y[0]
            ymin = y[0]
            ymax = y[-1]+dy

            z = kwargs.get('z', self.z/self.a0).flatten()
            dz = z[1] - z[0]
            zmin = z[0]
            zmax = z[-1]+dz

            X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
            
            for angle in [-2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3]:
                field_to_plot = theta.copy()
                field_to_plot[theta < angle - 1] = float('nan')
                field_to_plot[theta > angle + 1] = float('nan')
                field_to_plot[rho_normalized < 0.01] = float('nan')

                if np.nanmin(field_to_plot) < angle < np.nanmax(field_to_plot):

                    mesh, shadow_meshes = plot_surface_plotly(self,
                                            field_to_plot,
                                            value=angle,
                                            alpha=rho,
                                            color=plt_colormap_object((angle + np.pi) / (2 * np.pi)), 
                                            plt_colormap_object=plt_colormap_object,
                                            **kwargs
                                            )

                    fig.add_trace(mesh)

            theta = np.mod(theta, 2 * np.pi)

            field_to_plot = theta.copy()
            field_to_plot[theta < np.pi - 1] = float('nan')
            field_to_plot[theta > np.pi + 1] = float('nan')
            field_to_plot[rho_normalized < 0.01] = float('nan')

            if np.nanmin(field_to_plot) < np.pi < np.nanmax(field_to_plot):
                mesh, shadow_meshes = plot_surface_plotly(self,
                                            field_to_plot,
                                            value=np.pi,
                                            alpha=rho,
                                            color=plt_colormap_object((np.pi + np.pi) / (2 * np.pi)), 
                                            plt_colormap_object=plt_colormap_object,
                                            **kwargs
                                            )

                fig.add_trace(mesh)
        
        elif plot_method == 'phase_blob':
            
            phase_blob_threshold = kwargs.get('phase_blob_threshold', 0.5)

            if np.nanmin(rho_normalized)<phase_blob_threshold<np.nanmax(rho_normalized):

                mesh, shadow_meshes = plot_surface_plotly(self, 
                                        field=rho, 
                                        value=phase_blob_threshold*np.max(rho), 
                                        alpha=1,
                                        color=complex_field,
                                        plt_colormap_object=plt_colormap_object,
                                          **kwargs)

                fig.add_trace(mesh)

                if shadow_meshes is not None:
                    for shadow_mesh in shadow_meshes:
                        fig.add_trace(shadow_mesh)



    # if kwargs['colorbar'] and not(ax['colorbar']) and not(kwargs['field_is_nan']):
    #     ax['colormap_object'] = kwargs['colormap_object']
    #     fig.add_trace(tool_plotly_colorbar(ax, type='angle'))
    #     ax['colorbar'] = True


    if kwargs['colorbar'] and not(kwargs['field_is_nan']):
        ax['colormap_object'] = kwargs['colormap_object']
        fig.add_trace(tool_plotly_colorbar(ax, type='angle'))
        ax['colorbar'] = True

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_plotly(self, **kwargs)

    return fig, ax


from typing import TYPE_CHECKING, Any, Tuple, Union, List

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
import numpy as np

# Third-party library imports
import matplotlib.pyplot as plt
import matplotlib

# Local application imports
from comfit.tool import (
    tool_complete_field,
    tool_add_spacing_2D,
    tool_add_spacing_3D,
    tool_matplotlib_define_2D_plot_ax,
    tool_matplotlib_define_3D_plot_ax, 
    tool_set_plot_axis_properties_matplotlib
)

def plot_vector_field_matplotlib(
        self: 'BaseSystem',
        vector_field: Union[np.ndarray, List],
        **kwargs: Any
        ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plot the vector field using matplotlib.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    vector_field : np.ndarray
        Array containing the components of the vector field
    kwargs : Any
        spacing : int, optional. The spacing for the quiver plot which defaults 
        to self.xRes//20. Additional keyword arguments for plot customization, 
        see https://comfitlib.com/Plotting/

    Returns
    -------
    tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
        Figure and axes containing the plot

    Raises
    ------
    Exception
        If an invalid field is provided to the plot_vector_field function
    """

    spacing = kwargs.get('spacing', max(self.xRes//20,1))

    vector_field, fig, ax, kwargs = self.plot_prepare(vector_field, field_type = 'vector', **kwargs)

    # Extract coordinates
    x = kwargs.get('x', self.x/self.a0).flatten()
    dx = x[1] - x[0]
    xmin = x[0]
    xmax = x[-1]+dx
    
    if self.dim > 1:
        y = kwargs.get('y', self.y/self.a0).flatten()
        dy = y[1] - y[0]
        ymin = y[0]
        ymax = y[-1]+dy

    if self.dim > 2:
        z = kwargs.get('z', self.z/self.a0).flatten()
        dz = z[1] - z[0]
        zmin = z[0]
        zmax = z[-1]+dz


    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 1 ################
    ###############################################################
        
    if self.dim == 1 and vector_field.shape == (1,self.xRes):

        ax = tool_matplotlib_define_2D_plot_ax(fig, ax)

        X, Y = np.meshgrid(x, np.array([0]), indexing='ij')

        U = np.zeros(X.shape)
        V = np.zeros(X.shape)

        V[:,0] = vector_field[0]

        X,Y,U,V = tool_add_spacing_2D(X,Y,U,V,spacing)

        ax.quiver(X, Y, U, V, color='blue', angles='xy', scale_units='xy', scale=1)
        kwargs['ylim'] = [np.min(vector_field[0]), np.max(vector_field[0])]
        

    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 1 and vector_field.shape == (2,self.xRes):

        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True
            
        X, Y, Z = np.meshgrid(x, np.array([0]), np.array([0]), indexing='ij')

        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)

        V[:,0,0] = vector_field[0]
        W[:,0,0] = vector_field[1]

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        ax.quiver(X, Y, Z, U, V, W, color='blue')

        kwargs['ylim'] = [np.min(vector_field[0]), np.max(vector_field[0])]
        delta_y = kwargs['ylim'][1] - kwargs['ylim'][0]

        kwargs['zlim'] = [np.min(vector_field[1]), np.max(vector_field[1])]
        delta_z = kwargs['zlim'][1] - kwargs['zlim'][0]

        if delta_y < 0.15*delta_z:
            kwargs['ylim'] = [kwargs['ylim'][0] - 0.15*delta_z, kwargs['ylim'][1] + 0.15*delta_z]
        
        if delta_z < 0.15*delta_y:
            kwargs['zlim'] = [kwargs['zlim'][0] - 0.15*delta_y, kwargs['zlim'][1] + 0.15*delta_y]

    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 1 and vector_field.shape == (3,self.xRes):
        
        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True

        X, Y, Z = np.meshgrid(x, np.array([0]), np.array([0]), indexing='ij')

        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)

        # Add the vector field
        U[:,0,0] = vector_field[0]
        V[:,0,0] = vector_field[1]
        W[:,0,0] = vector_field[2]

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

        # Normalizing
        U = U/max_vector
        V = V/max_vector
        W = W/max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.size_x/15/self.a0)
        vy_scale = kwargs.get('vy_scale', 1)
        vz_scale = kwargs.get('vz_scale', 1)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        W = vz_scale*W

        ax.quiver(X, Y, Z, U, V, W, color='blue')

        kwargs['ylim'] = [-1,1]
        kwargs['zlim'] = [-1,1]

    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 1 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (1,self.xRes,self.yRes):
        
        ax = tool_matplotlib_define_2D_plot_ax(fig, ax)

        X, Y = np.meshgrid(x, y, indexing='ij')

        U = vector_field[0]
        V = np.zeros(X.shape)

        X,Y,U,V = tool_add_spacing_2D(X,Y,U,V,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2))
        
        U = U / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*dx)

        # Scaling
        U = vx_scale*U

        ax.quiver(X, Y, U, V, color='blue',
            angles='xy', scale_units='xy', scale=1,
            headwidth=3, headlength=4, headaxislength=3, pivot='middle')

    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (2,self.xRes,self.yRes):

        ax = tool_matplotlib_define_2D_plot_ax(fig, ax)

        X, Y = np.meshgrid(x, y, indexing='ij')

        X, Y, U, V = tool_add_spacing_2D(X,Y,vector_field[0],vector_field[1],spacing)

        max_vector = np.max(np.sqrt(U ** 2 + V ** 2))
    
        U = U / max_vector
        V = V / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', 1.0*spacing*dx)
        vy_scale = kwargs.get('vy_scale', 1.0*spacing*dy)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V

        ax.quiver(X, Y, U, V, color='blue', 
            angles='xy', scale_units='xy', scale=1,
            headwidth=3, headlength=4, headaxislength=3, pivot='middle')


    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (3,self.xRes,self.yRes):

        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True
        
        X, Y, Z = np.meshgrid(x, y, np.array([0]), indexing='ij')
        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)
        
        U[:,:,0] = vector_field[0]
        V[:,:,0] = vector_field[1]
        W[:,:,0] = vector_field[2]

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

        # Normalizing
        U = U / max_vector
        V = V / max_vector
        W = W / max_vector
        
        # Scale factors
        vx_scale = kwargs.get('vx_scale', 2*spacing*self.size_x/max_vector)
        vy_scale = kwargs.get('vy_scale', 2*spacing*self.size_y/max_vector)
        vz_scale = kwargs.get('vz_scale', spacing)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        W = vz_scale*W 
        
        ax.quiver(X, Y, Z, U, V, W, color='blue')
       
        kwargs['axis_equal'] = False
        kwargs['zlim'] = [-spacing,spacing]


    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 1 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (1,self.xRes,self.yRes,self.zRes):

        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True

        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')              

        # Define the vector field
        U = vector_field[0]
        V = np.zeros(U.shape)
        W = np.zeros(U.shape)

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2))
        U = U / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*dx)

        # Scaling
        U = vx_scale*U

        ax.quiver(X, Y, Z, U, V, W, color='blue')
        
    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (2,self.xRes,self.yRes,self.zRes):

        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True

        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

        # Define the vector field
        U = vector_field[0]
        V = vector_field[1]
        W = np.zeros(U.shape)

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2 + V ** 2))
        U = U / max_vector
        V = V / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*dx)
        vy_scale = kwargs.get('vy_scale', spacing*dy)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        

        ax.quiver(X, Y, Z, U, V, W, color='blue')
        
   
    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (3,self.xRes,self.yRes,self.zRes):

        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True

        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

        # Define the vector field
        U = vector_field[0]
        V = vector_field[1]
        W = vector_field[2]

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)



        

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

        U = U / max_vector
        V = V / max_vector
        W = W / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*dx)
        vy_scale = kwargs.get('vy_scale', spacing*dy)
        vz_scale = kwargs.get('vz_scale', spacing*dz)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        W = vz_scale*W
        
        ax.quiver(X, Y, Z, U, V, W, color='blue')
        

    ###############################################################
    ###########     NON-VALID DIMENSION            ################
    ###############################################################

    else:
        raise Exception("You have entered an invalid field to the plot_vector_field function.")

    kwargs['ax'] = ax
    tool_set_plot_axis_properties_matplotlib(self, **kwargs)
    return fig, ax
